SlotmapTicket

Struct SlotmapTicket 

pub struct SlotmapTicket { /* private fields */ }
Holds a ticket (think of it as a reference) to an element stored in the slotmap.

The ticket is used to access the element stored in the slotmap. The ticket is created when the element is inserted into the slotmap and is used to access the element until the element is removed from the slotmap.

Notice that the slotmap implementation ensures that each ticket is unique. If an element is removed from the slotmap, the ticket is invalidated and cannot be used, even if the slot is reused.

You can retrieve the element corresponding to the ticket using the LocklessSlotmap::get which will also guarantee that the element is not removed while it is being accessed. This ticket makes no such guarantees.