pub struct SlotmapEntry<'a, T> { /* private fields */ }Expand description
SlotmapEntry is a reference to an element stored in the slotmap.
The SlotmapEntry is created when the element is accessed in the slotmap. It ensures that the element cannot be removed while there is a thread actively accessing it.
Note
It is the responsibility of the user to ensure that the SlotmapEntry is dropped prior to removing the element from the slotmap. Failure to do so will result in a deadlock, where the erasing method will wait indefinitely for the SlotmapEntry to be dropped.
Example
use rust_lockless_slotmap::LocklessSlotmap;
use parking_lot::RawRwLock;
let slotmap: LocklessSlotmap<usize, RawRwLock> = LocklessSlotmap::new();
let ticket = slotmap.insert(42);
{
let entry = slotmap.get(ticket).unwrap();
assert_eq!(*entry, 42);
}Implementations
Sourceimpl<'a, T> SlotmapEntry<'a, T>
impl<'a, T> SlotmapEntry<'a, T>